from __future__ import annotations

import abc
import argparse
import dataclasses
import typing


class _BaseArgumentDefinition(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def add_argument(self, *, parser: argparse.ArgumentParser, field: dataclasses.Field[typing.Any], dest: str) -> None:
        raise NotImplementedError()


@dataclasses.dataclass(slots=True, kw_only=True)
class _ArgumentDefinition(_BaseArgumentDefinition):
    flags: tuple[str, ...] = ()
    required: bool = False
    help: str

    def add_argument(self, *, parser: argparse.ArgumentParser, field: dataclasses.Field[typing.Any], dest: str) -> None:
        if field.type == "bool":
            parser.add_argument(
                f"--{field.name}",
                *self.flags,
                default=field.default,
                dest=dest,
                action=argparse.BooleanOptionalAction,
                help=f"{self.help} (default: %(default)s)",
            )
        elif field.type.startswith("list["):
            parser.add_argument(
                f"--{field.name}",
                *self.flags,
                required=self.required,
                # not passing default (nor default_factor).
                # if argument isn't used, it will be None, and the
                # dataclass default is triggered
                dest=dest,
                action="append",
                help=f"{self.help}",
            )
        else:
            parser.add_argument(
                f"--{field.name}",
                *self.flags,
                required=self.required,
                default=field.default,
                dest=dest,
                help=f"{self.help}",
            )


def arg(*flags: str, required: bool = False, help: str) -> dict[typing.Any, typing.Any]:
    return {id(_BaseArgumentDefinition): _ArgumentDefinition(flags=flags, required=required, help=help)}


@dataclasses.dataclass(slots=True, kw_only=True)
class _ManualArgumentDefinition(_BaseArgumentDefinition):
    callback: typing.Callable[[argparse.ArgumentParser, str], None]

    def add_argument(self, *, parser: argparse.ArgumentParser, field: dataclasses.Field[typing.Any], dest: str) -> None:
        self.callback(parser, dest)


def manual(callback: typing.Callable[[argparse.ArgumentParser, str], None]) -> dict[typing.Any, typing.Any]:
    return {id(_BaseArgumentDefinition): _ManualArgumentDefinition(callback=callback)}


_TArgs = typing.TypeVar("_TArgs", bound="BaseArguments")


@dataclasses.dataclass(slots=True, kw_only=True)
class BaseArguments:
    def __post_init__(self) -> None:
        pass

    @classmethod
    def add_fields_to_parser(
        cls: type[_TArgs],
        parser: argparse.ArgumentParser,
        *,
        prefix: str = "",
    ) -> None:
        for field in dataclasses.fields(cls):
            argdef = field.metadata.get(id(_BaseArgumentDefinition), None)
            if argdef is None:
                continue
            assert isinstance(argdef, _BaseArgumentDefinition)
            dest = f"{prefix}{field.name}"
            argdef.add_argument(parser=parser, field=field, dest=dest)

    @classmethod
    def add_to_parser(
        cls: type[_TArgs],
        parser: argparse.ArgumentParser,
        *,
        prefix: str = "",
    ) -> ArgumentsParser[_TArgs]:
        cls.add_fields_to_parser(parser, prefix=prefix)
        return ArgumentsParser(cls=cls, prefix=prefix)


@dataclasses.dataclass(slots=True, kw_only=True)
class ArgumentsParser(typing.Generic[_TArgs]):
    cls: type[_TArgs]
    prefix: str

    def get_fields(self, args: argparse.Namespace) -> dict[str, typing.Any]:
        data = {}
        for field in dataclasses.fields(self.cls):
            argdef = field.metadata.get(id(_BaseArgumentDefinition), None)
            if argdef is None:
                continue
            value = getattr(args, f"{self.prefix}{field.name}")
            if not value is None:
                data[field.name] = value
        return data

    def from_args(self, args: argparse.Namespace) -> _TArgs:
        return self.cls(**self.get_fields(args))
